if not BGControlPublic then BGControlPublic = {} end
if not BG_Control_SV then BG_Control_SV = {} end
if not BG_Control_SV.Options then BG_Control_SV.Options = {} end

if not BG_Control_SV.TalentCache then BG_Control_SV.TalentCache = {} end
if not BG_Control_SV.MMRCache then BG_Control_SV.MMRCache = {} end
if not BG_Control_SV.NodeCache then BG_Control_SV.NodeCache = {} end

-- Local hidden frames, functions, and everything else
-- TODO: ADD LOCAL HERE
BGC = {
	frame = CreateFrame("Frame","BGC_Frame"),
	MAX_RADIUS = 120,
	ADDON_PREFIX = "BG_CONTROL",
	ADDON_MSG_SEPERATOR = ",",
	HEAL_SPELLS = 	{
			-- Swiftmend, Regrowth, Lifebloom
			["DRUID"]={["18562"]=true,["8936"]=true,["33763"]=true}, 
			-- Revival, Detonate, Uplift
			["MONK"]={["115310"]=true,["115460"]=true,["116670"]=true},
			-- Holy Shock,, Holy Radiance, Holy Light, Light of Dawn
			["PALADIN"]={["20473"]=true,["82327"]=true,["82326"]=true,["85222"]=true},
			-- Penance, Holy Nova | Binding HEal, Circle of Healing, Divine Hymn | Heal, Prayer of Healing
			["PRIEST"]={["47540"]=true,["132157"]=true,["32546"]=true,["34861"]=true,["64843"]=true,["2060"]=true,["596"]=true},
			-- Riptide, Chain Heal, Healing Wave
			["SHAMAN"]={["61295"]=true,["1064"]=true,["77472"]=true}
			},
	};

-- The "live" current BG Player Database for where people are located and who did what
-- TODO: Add Local HERE
PD = nil

function BGC.Player_Entering_Battleground(...)
	local _, type, _, _, _, _, _, mapID = GetInstanceInfo()
	if "pvp" == type then
		BGC.DebugPrint("Player has entered PVP Battleground ID:".. BGC.BGMapID());
		BGC.InitPD();
		BGC.frame:RegisterEvent("CHAT_MSG_ADDON");
		BGC.frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
		BGC.frame:RegisterEvent("PLAYER_LEAVING_WORLD");
		BGC.frame:RegisterEvent("GROUP_ROSTER_UPDATE");	
		BGC.frame:RegisterEvent("UPDATE_BATTLEFIELD_SCORE");
		BGC.frame:RegisterEvent("WORLD_MAP_UPDATE");
	end
		
end

-- Well, if we WERE in PVP... and we leave... we're out!
function BGC.Player_Leaving_World(...)
	BGC.DebugPrint("Player has left PVP Battlegrounds");
	BGC.frame:UnregisterEvent("CHAT_MSG_ADDON");
	BGC.frame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	BGC.frame:UnregisterEvent("PLAYER_LEAVING_WORLD");
	BGC.frame:UnregisterEvent("GROUP_ROSTER_UPDATE");
	BGC.frame:UnregisterEvent("UPDATE_BATTLEFIELD_SCORE");
	BGC.frame:UnregisterEvent("WORLD_MAP_UPDATE");
	
	-- TODO: Save off talent data to the saved variables
	BGC.DebugPrint("NOT clearing PD database!");
	--PD=nil
end


function BGC.BGMapID()
	local _, _, _, _, _, _, _, mapID = GetInstanceInfo()
	return mapID
	
end

function BGC.CheckNodeCache(mapID)
	if not SV.NodeCache[mapID] then
		SV.NodeCache[mapID] = {}
		local map = SV.NodeCache[mapID]
		BGC.InitNodeCache(map,mapID)
		BGC.CalculateNodeCacheRadius(map.node)
	end
end

function BGC.InitNodeCache(table,mapID)
	local index = 0
	table.node = {}
	local items = GetNumMapLandmarks()
	for i=1,items do
		local name, description, _, x, y, _, showInBattleMap = GetMapLandmarkInfo(i)

		-- If it's on the battle map, we care about it here too
		if showInBattleMap then
			index = index + 1
			if not table.node[index] then table.node[index] = {} end
			table.node[index].position = { x , y }
			table.node[index].landmarkID = i
		end
	
	end
	table.node["n"] = index
end

function BGC.CalculateNodeCacheRadius(map)
	local radius = BGC.MAX_RADIUS
	local mapSize = BGC.TableSize(map)
	for i=1, mapSize do
		for j=i,mapSize do
			radius = BGC.Distance(map[i].position,(map[j].position))
		end
	end
	map.nodeRadius = radius
end

function BGC.ClosestNodeToPlayer(player)
	-- TODO: Distance is NOT in yards!  Do we dynamically figure it out to account for new BGs?
	local closestNode,range
	local nodes = SV.NodeCache[BGC.BGMapID()]
	local spellRange = 0
	if player.range then spellRange = player.range end

	local mapSize = BGC.TableSize(nodes.node)

	for i=1, mapSize do
		local tmpRange = BGC.Distance(player.position,nodes.node[i].position)
		if not range or tmpRange < range then
			closestNode = nodes.node[i].landmarkID
			range = tmpRange
		end
	end
	return closestNode, range + spellRange <= BGC.nodeRadius ,range
end

function BGC.TableSize(table)
	return table.getn(table)
end

function BGC.Distance(from,to)
	-- TODO: this is NOT in yards...
	if not from or not to then
		BGC.DebugPrint("Distance failure: From("..tostring(from)..") To("..tostring(to)..")")
		return nil
	end
	
	local x = from[1] - to[1]
	local y = from[2] - to[2]
	return (x * x + y * y) ^ 0.5
end

function BGC.InitPD()
	PD={}
	PD.raidID = {}
	PD.damageMap = {}
	PD.healingMap = {}
	PD.NodeCount = {}
	PD.Player = {}
end


function BGC.Update_Battlefield_Score(...)
	--TODO: Use this to try to "prioritize" healers who are owning
end

function BGC.Combat_Log_Event_Unfiltered(...)
	local timeStamp, event, sourceName, sourceGUID, _, sourceFlags, destName, destGUID, _, destFlags, _, spellID = ...
	if BGC.IsPlayer(sourceFlags) and BGC.IsPlayer(destFlags) then
	
	
		if "SWING_DAMAGE" == event or "RANGE_DAMAGE" == event or "SPELL_DAMAGE" == event or "SPELL_HEAL" == event then
		if sourceGUID and not BGC.PlayerFromGUID(sourceGUID,sourceName,event).FullName then
			BGC.DebugPrint("Combat Log Fun: "..sourceName.." | "..sourceGUID)
		end
		
		if destGUID and not BGC.PlayerFromGUID(destGUID,destName,event).FullName then
			BGC.DebugPrint("Combat Log Fun: "..destName.." | "..destGUID)
		end
		end
	end
	BGC.ParseMessage(nil,event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID)
end

function BGC.ParseMessage(...)
	local isForign, event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags = ...
	-- TODO: Is it better to check flags BEFORE string compare? I assume so...
	if BGC.IsPlayer(sourceFlags) and BGC.IsPlayer(destFlags) then
		
		if "SWING_DAMAGE" == event then
			BGC.SwingDamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags, isForign)
		elseif "RANGE_DAMAGE" == event then
			local spellID= select(8,...)
			BGC.SpellDamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		elseif "SPELL_DAMAGE" == event then
			local spellID= select(8,...)
			BGC.SpellDamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		elseif "SPELL_HEAL" == event then
			local spellID= select(8,...)
			BGC.HealCast(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		end
	end
end

function BGC.Group_Roster_Update(...)
	BGC.DebugPrint("Modifying group size...")
	for i=1,40 do
		local raidID = "raid"..i;
		local name, realm = UnitName(raidID)
		if not realm or "" == realm then realm = GetRealmName() end
		PD.raidID[BGC.PlayerName(name,realm)] = raidID
	end
end

function BGC.SwingDamageDone(event,time,from,fromFlags,to,toFlags,isForign,fN,tN)
	local fromName, toName

	if BGC.IsHostile(fromFlags) then
		fromName = BGC.PlayerFromGUID(from).FullName
		toName = BGC.PlayerFromGUID(to).FullName
	elseif BGC.IsHostile(toFlags) then
		fromName = BGC.PlayerFromGUID(to).FullName
		toName = BGC.PlayerFromGUID(from).FullName
	end

	if not PD.damageMap[fromName] then PD.damageMap[fromName] = {} end
	local map = PD.damageMap[fromName]
	-- This is an update (used when doing Addon cross-talk)
	if not map.time or map.time < time then	
		local x, y = UnitPosition(BGC.RaidID(toName))
		map.time = time
		map.range = 0
		map.target = toName
		map.position = { x , y }
		BGC.UpdateDamage(map)
	end

	if not isForign then
		BGC.SendMessage(event,time,from,fromFlags,to,toFlags)
	end
end

function BGC.SendMessage(...)
	-- TODO: Make this better... have a function for damage, and a function for "other"
	--local msg = table.concat(table.pack(...),BGC.ADDON_MSG_SEPERATOR)
	--SendAddonMessage(BGC.ADDON_PREFIX, msg "BATTLEGROUND")
end

function BGC.Chat_Msg_Addon(...)
	local prefix, message = ...
	-- I could check the channel, but frankly I don't care... unless someone
	--	is trying to screw with this data, it's going to be fine
	if prefix == BGC.ADDON_PREFIX then
		BGC.ParseMessage(true,table.unpack(split(message,BGC.ADDON_MSG_SEPERATOR)))
	end
end

function BGC.SpellDamageDone(event,time,fromGUID,fromFlags,toGUID,toFlags,spellID, isForign)	
	local fromName, toName

	if BGC.IsHostile(fromFlags) then
		fromName = BGC.PlayerFromGUID(fromGUID).FullName
		toName = BGC.PlayerFromGUID(toGUID).FullName
	elseif BGC.IsHostile(toFlags) then
		fromName = BGC.PlayerFromGUID(toGUID).FullName
		toName = BGC.PlayerFromGUID(fromGUID).FullName
	end

	if not PD.damageMap[fromName] then PD.damageMap[fromName] = {} end
	local map = PD.damageMap[fromName]
	
	-- This is an update (used when doing Addon cross-talk)
	if not map.time or map.time < time then
		map.time = time
		map.range = BGC.SpellRange(spellID);
		map.target = toName
		if not BGC.RaidID(toName) then
			BGC.DebugPrint("Trying to unitPosition with name: "..toName.." and guid "..toGUID)
		end
		local x, y = UnitPosition(BGC.RaidID(toName))
		map.position = { x , y }
		BGC.UpdateDamage(map)
	end

	if not isForign then
		BGC.SendMessage(event,time,from,fromFlags,to,toFlags,spellID)
	end
end

function BGC.HealCast(event, time,from,fromFlags,to,toFlags,spellID, isForign)
	-- We don't care about actions for our players... only enemy
	--	Also, self-healing gives us NO data... ignore it
	if BGC.IsHostile(fromFlags) and from ~= to then
		local fromPlayer = BGC.PlayerFromGUID(from)

		-- If this person CAN be a healer, and they aren't flagged as one
		--	check the spell for a heal!
		-- TODO: When we know what talent spec they are, stop...
		if fromPlayer.CanBeHealer and not fromPlayer.IsHealer then
			fromPlayer.IsHealer = BGC.IsHealSpell(spellID,fromPlayer.Race)
		end

		local fromName = fromPlayer.FullName
		if not PD.healingMap[fromName] then PD.healingMap[fromName] = {} end
		local map = PD.healingMap[fromName]
		
		-- This is an update (used when doing Addon cross-talk)
		if not map.time or map.time < time then
			
			local toName = BGC.PlayerFromGUID(to).Name
			map.time = time
			map.range = BGC.SpellRange(spellID);
			map.target=toName;
			BGC.UpdateHealer(map)
		end

		-- If it's not from me... someone else got it too, so don't forward it
		if not isForign then
			BGC.SendMessage(event, time,from,fromFlags,to,toFlags,spellID)
		end
	end
end

function BGC.RaidID(name)
	if not PD.raidID[name] then
		--BGC.DebugPrint("Bad RaidID: "..name)
	end
	return PD.raidID[name]
end

function BGC.UpdateHealer(healingMap)
	local dm = PD.damageMap[healingMap.target]
	
	-- If we don't have a damage entry for this target... something's off
	if dm then
		local nodeID,isAtNode = BGC.ClosestNodeToPlayer(dm)
		BGC.UpdatePlayer(nodeID,isAtNode,healingMap)
	end
	
end

function BGC.UpdateDamage(damageMap)
	local nodeID,isAtNode = BGC.ClosestNodeToPlayer(damageMap)
	BGC.UpdatePlayer(nodeID,isAtNode,damageMap)
end

function BGC.UpdatePlayer(nodeID,isAtNode,player)
	if not PD.NodeCount[nodeID] then PD.NodeCount[nodeID] = {} end
	local map = PD.NodeCounts[nodeID]

	if player.lastNode then
		local nodeCounts = PD.NodeCount[player.lastNode.id]
		if player.lastNode.close then
			nodeCount.count = nodeCount.count - 1
		else
			nodeCount.nearby = nodeCount.nearby - 1
		end
	end 

	if player.node then
		player.lastNode = player.node
	end
 
	if isAtNode then
		map.count = map.count + 1
	else
		map.nearby = map.nearby + 1
	end

	local DEBUG= SV.NodeCache[BGC.GetMapID()][nodeID]
	BGC.DebugPrint(DEBUG.name .. ": ".. map.count .. " with " .. map.nearby .. " nearby.")

	-- TODO: Update the onhover/frame
end

-- TODO: Timer tick to "stale" old data

function BGC.PlayerName(name,realm)
	return name.."-"..realm
end

function BGC.SpellRange(id)
	local _, _, _, _, _, maxRange = GetSpellInfo(id)
	return maxRange
end

-- We need to do this, name doesn't include own realm and
--	guid is really long to compare... 
function BGC.PlayerFromGUID(guid,name,event)
	local m = PD.Player[guid]
	if not m then 
		PD.Player[guid] = {}
		m = PD.Player[guid]
		local _,class,_,race,sex,name,realm = GetPlayerInfoByGUID(guid)
		if not realm or "" == realm then realm = GetRealmName() end
		m.Class = class
		m.Race = race
		m.Sex = sex
		m.Name = name
		m.Realm = realm
		m.FullName = BGC.PlayerName(name,realm)
		m.CanBeHealer = BGC.CanBeHealer(race)
		m.IsHealer = nil
		if guid and not name then 
			BGC.DebugPrint("GetPlayerInfoByGUID failure "..tostring(guid).." | "..tostring(name).." | "..tostring(event).." | "..realm)
		end
	end
end

function BGC.IsHostile(flags)
	return bit.band(flags, COMBATLOG_OBJECT_REACTION_HOSTILE) > 0	
end

function BGC.IsPlayer(flags)	
	return bit.band(flags,COMBATLOG_OBJECT_TYPE_PLAYER) > 0
end

function BGC.Addon_Loaded(...)
	local name = ...
	if "BG_Control" == name then
		BGC.DebugPrint("Loaded BG_Control!")
	
	
		-- TODO: DELETE ME SOMETIME
		-- DEBUG LINEs
		BGC.DebugPrint("Deleting previous database for debugging")
		BG_Control_SV = {}
		BG_Control_SV.Options = {}
		BG_Control_SV.TalentCache = {}
		BG_Control_SV.MMRCache = {}
		BG_Control_SV.NodeCache = {}	


		SV = BG_Control_SV		

		-- Get rid of this garbage... never used again
		BGC.frame:UnregisterEvent("ADDON_LOADED");
		BGC.Addon_Loaded = nil
	end
end

function BGC.OnEvent(self, event, ...)
	-- Fires more than one a second
	if "COMBAT_LOG_EVENT_UNFILTERED" == event then
		BGC.Combat_Log_Event_Unfiltered(...);

	-- These are filtered combat logs, so they happen less frequently
	elseif "CHAT_MSG_ADDON" == event then
		BGC.Chat_Msg_Addon(...);

	-- These updates happen semi-often on kills/point changes
	elseif "UPDATE_BATTLEFIELD_SCORE" == event then
		BGC.Update_Battlefield_Score(...);

	-- This fires only when someone leaves or joins, hopefully not often
	elseif "GROUP_ROSTER_UPDATE" == event then
		BGC.Group_Roster_Update(...);

	-- These fire once a BG sessions
	elseif "PLAYER_ENTERING_BATTLEGROUND" == event then
		BGC.Player_Entering_Battleground(...);
	elseif "PLAYER_LEAVING_WORLD" == event then
		BGC.Player_Leaving_World(...);

	-- These things only happen "once" ever and then are unregistered
	elseif "WORLD_MAP_UPDATE" == event then
		BGC.World_Map_Update(...);
	elseif "ADDON_LOADED" == event then
		BGC.Addon_Loaded(...);
	end
end

function BGC.DebugPrint(str)
	-- This contains a curseforge macro to not package when deployed.
	--	See repository
	
end

-- Node Info is available only AFTER this event is fired
function BGC.World_Map_Update(...)
	BGC.DebugPrint("Updating World Map")
	BGC.CheckNodeCache(BGC.BGMapID())
	--BGC.Instance_Group_Size_Changed() -- This is in case of a reload
end


BGC.frame:SetScript("OnEvent",BGC.OnEvent);
BGC.frame:RegisterEvent("ADDON_LOADED");
BGC.frame:RegisterEvent("PLAYER_ENTERING_BATTLEGROUND");


function BGC.CanBeHealer(race)
	return BGC.HEAL_SPELLS[race]
end

function BGC.IsHealSpell(spellID,race)
	if not race then
		for i=1,GetNumClasses() do
			local _,r = GetClassInfo(i)
			if BGC.IsHealSpell(spellID,r) then
				return true
			end
		end
		return nil
	end

	if not BGC.CanBeHealer(race) then return nil end

	return BGC.HEAL_SPELLS[race][spellID]
end

-- Druid
	-- Swiftmend - 18562
	-- Regrowth - 8936
	-- Lifebloom - 33763 -- HOT

-- Monk
	-- REvival - 115310
	-- Detonate Chi - 115460
	-- Uplift - 116670

-- Paladin
	-- Holy Shock - 20473
	-- Holy Radiance - 82327
	-- Holy Light - 82326
	-- Light of Dawn - 85222

-- Priest
	-- Disc
		-- Penance - 47540
		--Holy Nova - 132157
	-- Holy
		-- Binding Heal - 32546
		-- Circle of Healing - 34861
		-- Divine Hymn - 64843
	-- Both
		-- Heal - 2060
		-- Prayer of Healing - 596

-- Shaman
	-- Riptide - 61295
	-- Chain Heal - 1064
	-- Healing Wave - 77472
